@@ -16,6 +16,7 @@
 
 # A tool to run tests in many different ways.
 
+import shlex
 import subprocess, sys, os, argparse
 import pickle
 from mesonbuild import build
@@ -25,6 +26,7 @@ import time, datetime, multiprocessing, json
 import concurrent.futures as conc
 import platform
 import signal
+import random
 
 # GNU autotools interprets a return code of 77 from tests it executes to
 # mean that the test should be skipped.
@@ -60,16 +62,18 @@ parser.add_argument('--gdb', default=False, dest='gdb', action='store_true',
                     help='Run test under gdb.')
 parser.add_argument('--list', default=False, dest='list', action='store_true',
                     help='List available tests.')
-parser.add_argument('--wrapper', default=None, dest='wrapper',
+parser.add_argument('--wrapper', default=None, dest='wrapper', type=shlex.split,
                     help='wrapper to run tests with (e.g. Valgrind)')
 parser.add_argument('-C', default='.', dest='wd',
                     help='directory to cd into before running')
-parser.add_argument('--suite', default=None, dest='suite',
+parser.add_argument('--suite', default=[], dest='include_suites', action='append', metavar='SUITE',
                     help='Only run tests belonging to the given suite.')
+parser.add_argument('--no-suite', default=[], dest='exclude_suites', action='append', metavar='SUITE',
+                    help='Do not run tests belonging to the given suite.')
 parser.add_argument('--no-stdsplit', default=True, dest='split', action='store_false',
                     help='Do not split stderr and stdout in test logs.')
 parser.add_argument('--print-errorlogs', default=False, action='store_true',
-                    help="Whether to print faling tests' logs.")
+                    help="Whether to print failing tests' logs.")
 parser.add_argument('--benchmark', default=False, action='store_true',
                     help="Run benchmarks instead of tests.")
 parser.add_argument('--logbase', default='testlog',
@@ -78,13 +82,20 @@ parser.add_argument('--num-processes', default=determine_worker_count(), type=in
                     help='How many parallel processes to use.')
 parser.add_argument('-v', '--verbose', default=False, action='store_true',
                     help='Do not redirect stdout and stderr')
-parser.add_argument('-t', '--timeout-multiplier', type=float, default=1.0,
+parser.add_argument('-q', '--quiet', default=False, action='store_true',
+                    help='Produce less output to the terminal.')
+parser.add_argument('-t', '--timeout-multiplier', type=float, default=None,
                     help='Define a multiplier for test timeout, for example '
                     ' when running tests in particular conditions they might take'
                     ' more time to execute.')
-parser.add_argument('args', nargs='*')
+parser.add_argument('--setup', default=None, dest='setup',
+                    help='Which test setup to use.')
+parser.add_argument('--test-args', default=[], type=shlex.split,
+                    help='Arguments to pass to the specified test(s) or all tests')
+parser.add_argument('args', nargs='*',
+                    help='Optional list of tests to run')
 
-class TestRun():
+class TestRun:
     def __init__(self, res, returncode, should_fail, duration, stdo, stde, cmd,
                  env):
         self.res = res
@@ -101,9 +112,7 @@ class TestRun():
         if self.cmd is None:
             res += 'NONE\n'
         else:
-            res += "\n%s %s\n" %(' '.join(
-                ["%s='%s'" % (k, v) for k, v in self.env.items()]),
-                ' ' .join(self.cmd))
+            res += "%s%s\n" % (''.join(["%s='%s' " % (k, v) for k, v in self.env.items()]), ' ' .join(self.cmd))
         if self.stdo:
             res += '--- stdout ---\n'
             res += self.stdo
@@ -126,13 +135,12 @@ def decode(stream):
         return stream.decode('iso-8859-1', errors='ignore')
 
 def write_json_log(jsonlogfile, test_name, result):
-    jresult = {'name' : test_name,
-              'stdout' : result.stdo,
-              'result' : result.res,
-              'duration' : result.duration,
-              'returncode' : result.returncode,
-              'command' : result.cmd,
-              }
+    jresult = {'name': test_name,
+               'stdout': result.stdo,
+               'result': result.res,
+               'duration': result.duration,
+               'returncode': result.returncode,
+               'command': result.cmd}
     if isinstance(result.env, dict):
         jresult['env'] = result.env
     else:
@@ -150,38 +158,19 @@ class TestHarness:
     def __init__(self, options):
         self.options = options
         self.collected_logs = []
-        self.failed_tests = []
-        self.error_count = 0
+        self.fail_count = 0
+        self.success_count = 0
+        self.skip_count = 0
+        self.timeout_count = 0
         self.is_run = False
-        self.cant_rebuild = False
+        self.tests = None
+        self.suites = None
         if self.options.benchmark:
-            self.datafile = os.path.join(options.wd, 'meson-private/meson_benchmark_setup.dat')
+            self.load_datafile(os.path.join(options.wd, 'meson-private', 'meson_benchmark_setup.dat'))
         else:
-            self.datafile = os.path.join(options.wd, 'meson-private/meson_test_setup.dat')
-
-    def rebuild_all(self):
-        if not os.path.isfile(os.path.join(self.options.wd, 'build.ninja')):
-            print("Only ninja backend is supported to rebuilt tests before running them.")
-            self.cant_rebuild = True
-            return True
-
-        ninja = environment.detect_ninja()
-        if not ninja:
-            print("Can't find ninja, can't rebuild test.")
-            self.cant_rebuild = True
-            return False
-
-        p = subprocess.Popen([ninja, '-C', self.options.wd])
-        (stdo, stde) = p.communicate()
-
-        if p.returncode != 0:
-            print("Could not rebuild")
-            return False
-
-        return True
+            self.load_datafile(os.path.join(options.wd, 'meson-private', 'meson_test_setup.dat'))
 
     def run_single_test(self, wrap, test):
-        failling = False
         if test.fname[0].endswith('.jar'):
             cmd = ['java', '-jar'] + test.fname
         elif not test.is_cross and run_with_mono(test.fname[0]):
@@ -204,15 +193,24 @@ class TestHarness:
             stde = None
             returncode = GNU_SKIP_RETURNCODE
         else:
-            cmd = wrap + cmd + test.cmd_args
+            cmd = wrap + cmd + test.cmd_args + self.options.test_args
             starttime = time.time()
             child_env = os.environ.copy()
+            child_env.update(self.options.global_env.get_env(child_env))
             if isinstance(test.env, build.EnvironmentVariables):
                 test.env = test.env.get_env(child_env)
 
             child_env.update(test.env)
             if len(test.extra_paths) > 0:
-                child_env['PATH'] = child_env['PATH'] + ';'.join([''] + test.extra_paths)
+                child_env['PATH'] += ';'.join([''] + test.extra_paths)
+
+            # If MALLOC_PERTURB_ is not set, or if it is set to an empty value,
+            # (i.e., the test or the environment don't explicitly set it), set
+            # it ourselves. We do this unconditionally because it is extremely
+            # useful to have in tests.
+            # Setting MALLOC_PERTURB_="0" will completely disable this feature.
+            if 'MALLOC_PERTURB_' not in child_env or not child_env['MALLOC_PERTURB_']:
+                child_env['MALLOC_PERTURB_'] = str(random.randint(1, 255))
 
             setsid = None
             stdout = None
@@ -256,39 +254,66 @@ class TestHarness:
                 stde = decode(stde)
             if timed_out:
                 res = 'TIMEOUT'
-                failling = True
-            if p.returncode == GNU_SKIP_RETURNCODE:
+                self.timeout_count += 1
+                self.fail_count += 1
+            elif p.returncode == GNU_SKIP_RETURNCODE:
                 res = 'SKIP'
-            elif (not test.should_fail and p.returncode == 0) or \
-                (test.should_fail and p.returncode != 0):
+                self.skip_count += 1
+            elif test.should_fail == bool(p.returncode):
                 res = 'OK'
+                self.success_count += 1
             else:
                 res = 'FAIL'
-                failling = True
+                self.fail_count += 1
             returncode = p.returncode
         result = TestRun(res, returncode, test.should_fail, duration, stdo, stde, cmd, test.env)
 
-        if failling:
-            self.failed_tests.append(result)
-
         return result
 
     def print_stats(self, numlen, tests, name, result, i, logfile, jsonlogfile):
-        startpad = ' '*(numlen - len('%d' % (i+1)))
-        num = '%s%d/%d' % (startpad, i+1, len(tests))
-        padding1 = ' '*(38-len(name))
-        padding2 = ' '*(8-len(result.res))
+        startpad = ' ' * (numlen - len('%d' % (i + 1)))
+        num = '%s%d/%d' % (startpad, i + 1, len(tests))
+        padding1 = ' ' * (38 - len(name))
+        padding2 = ' ' * (8 - len(result.res))
         result_str = '%s %s  %s%s%s%5.2f s' % \
             (num, name, padding1, result.res, padding2, result.duration)
-        print(result_str)
+        if not self.options.quiet or result.res != 'OK':
+            print(result_str)
         result_str += "\n\n" + result.get_log()
-        if (result.returncode != GNU_SKIP_RETURNCODE) and \
-            (result.returncode != 0) != result.should_fail:
-            self.error_count += 1
+        if (result.returncode != GNU_SKIP_RETURNCODE) \
+                and (result.returncode != 0) != result.should_fail:
             if self.options.print_errorlogs:
                 self.collected_logs.append(result_str)
-        logfile.write(result_str)
-        write_json_log(jsonlogfile, name, result)
+        if logfile:
+            logfile.write(result_str)
+        if jsonlogfile:
+            write_json_log(jsonlogfile, name, result)
+
+    def print_summary(self, logfile, jsonlogfile):
+        msg = '''
+OK:      %4d
+FAIL:    %4d
+SKIP:    %4d
+TIMEOUT: %4d
+''' % (self.success_count, self.fail_count, self.skip_count, self.timeout_count)
+        print(msg)
+        if logfile:
+            logfile.write(msg)
+
+    def print_collected_logs(self):
+        if len(self.collected_logs) > 0:
+            if len(self.collected_logs) > 10:
+                print('\nThe output from 10 first failed tests:\n')
+            else:
+                print('\nThe output from the failed tests:\n')
+            for log in self.collected_logs[:10]:
+                lines = log.splitlines()
+                if len(lines) > 104:
+                    print('\n'.join(lines[0:4]))
+                    print('--- Listing only the last 100 lines from a long log. ---')
+                    lines = lines[-100:]
+                for line in lines:
+                    print(line)
 
     def doit(self):
         if self.is_run:
@@ -297,179 +322,270 @@ class TestHarness:
             print('Test data file. Probably this means that you did not run this in the build directory.')
             return 1
         self.is_run = True
-        logfilename = self.run_tests(self.options.logbase)
-        if len(self.collected_logs) > 0:
-            if len(self.collected_logs) > 10:
-                print('\nThe output from 10 first failed tests:\n')
-            else:
-                print('\nThe output from the failed tests:\n')
-            for log in self.collected_logs[:10]:
-                lines = log.splitlines()
-                if len(lines) > 100:
-                    print(lines[0])
-                    print('--- Listing only the last 100 lines from a long log. ---')
-                    lines = lines[-99:]
-                for line in lines:
-                    print(line)
-        print('Full log written to %s.' % logfilename)
-        return self.error_count
+        tests = self.get_tests()
+        if not tests:
+            return 0
+        self.run_tests(tests)
+        return self.fail_count
+
+    @staticmethod
+    def split_suite_string(suite):
+        if ':' in suite:
+            return suite.split(':', 1)
+        else:
+            return suite, ""
+
+    @staticmethod
+    def test_in_suites(test, suites):
+        for suite in suites:
+            (prj_match, st_match) = TestHarness.split_suite_string(suite)
+            for prjst in test.suite:
+                (prj, st) = TestHarness.split_suite_string(prjst)
+                if prj_match and prj != prj_match:
+                    continue
+                if st_match and st != st_match:
+                    continue
+                return True
+        return False
+
+    def test_suitable(self, test):
+        return (len(self.options.include_suites) == 0 or TestHarness.test_in_suites(test, self.options.include_suites)) \
+            and not TestHarness.test_in_suites(test, self.options.exclude_suites)
+
+    def load_suites(self):
+        ss = set()
+        for t in self.tests:
+            for s in t.suite:
+                ss.add(s)
+        self.suites = list(ss)
+
+    def load_tests(self):
+        with open(self.datafile, 'rb') as f:
+            self.tests = pickle.load(f)
+
+    def load_datafile(self, datafile):
+        self.datafile = datafile
+        self.load_tests()
+        self.load_suites()
 
     def get_tests(self):
-        with open(self.datafile, 'rb') as f:
-            tests = pickle.load(f)
+        if not self.tests:
+            print('No tests defined.')
+            return []
+
+        if len(self.options.include_suites) or len(self.options.exclude_suites):
+            tests = []
+            for tst in self.tests:
+                if self.test_suitable(tst):
+                    tests.append(tst)
+        else:
+            tests = self.tests
+
+        if self.options.args:
+            tests = [t for t in tests if t.name in self.options.args]
+
+        if not tests:
+            print('No suitable tests defined.')
+            return []
+
         for test in tests:
             test.rebuilt = False
 
         return tests
 
-    def run_tests(self, log_base):
-        logfile_base = os.path.join(self.options.wd, 'meson-logs', log_base)
-        if self.options.wrapper is None:
-            wrap = []
-            logfilename = logfile_base + '.txt'
-            jsonlogfilename = logfile_base+ '.json'
-        else:
-            wrap = self.options.wrapper.split()
-            namebase = wrap[0]
-            logfilename = logfile_base + '-' + namebase.replace(' ', '_') + '.txt'
-            jsonlogfilename = logfile_base + '-' + namebase.replace(' ', '_') + '.json'
-        tests = self.get_tests()
-        if len(tests) == 0:
-            print('No tests defined.')
-            return
-        numlen = len('%d' % len(tests))
-        executor = conc.ThreadPoolExecutor(max_workers=self.options.num_processes)
-        futures = []
-        filtered_tests = filter_tests(self.options.suite, tests)
+    def open_log_files(self):
+        if not self.options.logbase or self.options.verbose:
+            return None, None, None, None
 
-        jsonlogfile = None
-        logfile = None
+        namebase = None
+        logfile_base = os.path.join(self.options.wd, 'meson-logs', self.options.logbase)
+
+        if self.options.wrapper:
+            namebase = os.path.split(self.get_wrapper()[0])[1]
+        elif self.options.setup:
+            namebase = self.options.setup
+
+        if namebase:
+            logfile_base += '-' + namebase.replace(' ', '_')
+        logfilename = logfile_base + '.txt'
+        jsonlogfilename = logfile_base + '.json'
+
+        jsonlogfile = open(jsonlogfilename, 'w')
+        logfile = open(logfilename, 'w')
+
+        logfile.write('Log of Meson test suite run on %s\n\n'
+                      % datetime.datetime.now().isoformat())
+
+        return logfile, logfilename, jsonlogfile, jsonlogfilename
+
+    def get_wrapper(self):
+        wrap = []
+        if self.options.gdb:
+            wrap = ['gdb', '--quiet', '--nh']
+            if self.options.repeat > 1:
+                wrap += ['-ex', 'run', '-ex', 'quit']
+            # Signal the end of arguments to gdb
+            wrap += ['--args']
+        if self.options.wrapper:
+            wrap += self.options.wrapper
+        assert(isinstance(wrap, list))
+        return wrap
+
+    def get_pretty_suite(self, test, tests):
+        if len(self.suites) > 1:
+            rv = TestHarness.split_suite_string(test.suite[0])[0]
+            s = "+".join(TestHarness.split_suite_string(s)[1] for s in test.suite)
+            if len(s):
+                rv += ":"
+            return rv + s + " / " + test.name
+        else:
+            return test.name
+
+    def run_tests(self, tests):
         try:
-            if not self.options.verbose:
-                jsonlogfile =  open(jsonlogfilename, 'w')
-                logfile = open(logfilename, 'w')
-                logfile.write('Log of Meson test suite run on %s.\n\n' %
-                            datetime.datetime.now().isoformat())
+            executor = None
+            logfile = None
+            jsonlogfile = None
+            futures = []
+            numlen = len('%d' % len(tests))
+            (logfile, logfilename, jsonlogfile, jsonlogfilename) = self.open_log_files()
+            wrap = self.get_wrapper()
 
             for i in range(self.options.repeat):
-                for i, test in enumerate(filtered_tests):
-                    if test.suite[0] == '':
-                        visible_name = test.name
-                    else:
-                        if self.options.suite is not None:
-                            visible_name = self.options.suite + ' / ' + test.name
-                        else:
-                            visible_name = test.suite[0] + ' / ' + test.name
+                for i, test in enumerate(tests):
+                    visible_name = self.get_pretty_suite(test, tests)
 
-                    if not test.is_parallel:
-                        self.drain_futures(futures)
+                    if self.options.gdb:
+                        test.timeout = None
+
+                    if not test.is_parallel or self.options.gdb:
+                        self.drain_futures(futures, logfile, jsonlogfile)
                         futures = []
                         res = self.run_single_test(wrap, test)
-                        if not self.options.verbose:
-                            self.print_stats(numlen, filtered_tests, visible_name, res, i,
-                                            logfile, jsonlogfile)
+                        self.print_stats(numlen, tests, visible_name, res, i, logfile, jsonlogfile)
                     else:
+                        if not executor:
+                            executor = conc.ThreadPoolExecutor(max_workers=self.options.num_processes)
                         f = executor.submit(self.run_single_test, wrap, test)
-                        if not self.options.verbose:
-                            futures.append((f, numlen, filtered_tests, visible_name, i,
-                                            logfile, jsonlogfile))
+                        futures.append((f, numlen, tests, visible_name, i, logfile, jsonlogfile))
+                    if self.options.repeat > 1 and self.fail_count:
+                        break
+                if self.options.repeat > 1 and self.fail_count:
+                    break
+
             self.drain_futures(futures, logfile, jsonlogfile)
+            self.print_summary(logfile, jsonlogfile)
+            self.print_collected_logs()
+
+            if logfilename:
+                print('Full log written to %s' % logfilename)
         finally:
             if jsonlogfile:
                 jsonlogfile.close()
             if logfile:
                 logfile.close()
 
-        return logfilename
-
-
     def drain_futures(self, futures, logfile, jsonlogfile):
         for i in futures:
             (result, numlen, tests, name, i, logfile, jsonlogfile) = i
-            if self.options.repeat > 1 and self.failed_tests:
+            if self.options.repeat > 1 and self.fail_count:
                 result.cancel()
-            elif not self.options.verbose:
-                self.print_stats(numlen, tests, name, result.result(), i, logfile, jsonlogfile)
-            else:
+            if self.options.verbose:
                 result.result()
-
-        if self.options.repeat > 1 and self.failed_tests:
-            if not self.options.verbose:
-                for res in self.failed_tests:
-                    print('Test failed:\n\n-- stdout --\n')
-                    print(res.stdo)
-                    print('\n-- stderr --\n')
-                    print(res.stde)
-                    return 1
-
-            return
+            self.print_stats(numlen, tests, name, result.result(), i, logfile, jsonlogfile)
 
     def run_special(self):
         'Tests run by the user, usually something like "under gdb 1000 times".'
         if self.is_run:
             raise RuntimeError('Can not use run_special after a full run.')
-        if self.options.wrapper is not None:
-            wrap = self.options.wrapper.split(' ')
-        else:
-            wrap = []
-        if self.options.gdb and len(wrap) > 0:
-            print('Can not specify both a wrapper and gdb.')
-            return 1
         if os.path.isfile('build.ninja'):
             subprocess.check_call([environment.detect_ninja(), 'all'])
         tests = self.get_tests()
-        if self.options.list:
-            for i in tests:
-                print(i.name)
+        if not tests:
             return 0
-        for t in tests:
-            if t.name in self.options.args:
-                for i in range(self.options.repeat):
-                    print('Running: %s %d/%d' % (t.name, i+1, self.options.repeat))
-                    if self.options.gdb:
-                        wrap = ['gdb', '--quiet']
-                        if len(t.cmd_args) > 0:
-                            wrap.append('--args')
-                        if self.options.repeat > 1:
-                            # The user wants to debug interactively, so no timeout.
-                            t.timeout = None
-                            wrap += ['-ex', 'run', '-ex', 'quit']
+        self.run_tests(tests)
+        return self.fail_count
 
-                        res = self.run_single_test(wrap, t)
-                    else:
-                        res = self.run_single_test(wrap, t)
-                        if (res.returncode == 0 and res.should_fail) or \
-                                (res.returncode != 0 and not res.should_fail):
-                            if not self.options.verbose:
-                                print('Test failed:\n\n-- stdout --\n')
-                                print(res.stdo)
-                                print('\n-- stderr --\n')
-                                print(res.stde)
-                            return 1
-        return 0
 
-def filter_tests(suite, tests):
-    if suite is None:
-        return tests
-    return [x for x in tests if suite in x.suite]
+def list_tests(th):
+    tests = th.get_tests()
+    for t in tests:
+        print(th.get_pretty_suite(t, tests))
 
+def merge_suite_options(options):
+    buildfile = os.path.join(options.wd, 'meson-private/build.dat')
+    with open(buildfile, 'rb') as f:
+        build = pickle.load(f)
+    setups = build.test_setups
+    if options.setup not in setups:
+        sys.exit('Unknown test setup: %s' % options.setup)
+    current = setups[options.setup]
+    if not options.gdb:
+        options.gdb = current.gdb
+    if options.timeout_multiplier is None:
+        options.timeout_multiplier = current.timeout_multiplier
+#    if options.env is None:
+#        options.env = current.env # FIXME, should probably merge options here.
+    if options.wrapper is not None and current.exe_wrapper is not None:
+        sys.exit('Conflict: both test setup and command line specify an exe wrapper.')
+    if options.wrapper is None:
+        options.wrapper = current.exe_wrapper
+    return current.env
+
+def rebuild_all(wd):
+    if not os.path.isfile(os.path.join(wd, 'build.ninja')):
+        print("Only ninja backend is supported to rebuild tests before running them.")
+        return True
+
+    ninja = environment.detect_ninja()
+    if not ninja:
+        print("Can't find ninja, can't rebuild test.")
+        return False
+
+    p = subprocess.Popen([ninja, '-C', wd])
+    (stdo, stde) = p.communicate()
+
+    if p.returncode != 0:
+        print("Could not rebuild")
+        return False
+
+    return True
 
 def run(args):
     options = parser.parse_args(args)
+
     if options.benchmark:
         options.num_processes = 1
 
+    if options.setup is not None:
+        global_env = merge_suite_options(options)
+    else:
+        global_env = build.EnvironmentVariables()
+        if options.timeout_multiplier is None:
+            options.timeout_multiplier = 1
+
+    setattr(options, 'global_env', global_env)
+
+    if options.verbose and options.quiet:
+        print('Can not be both quiet and verbose at the same time.')
+        return 1
+
     if options.gdb:
         options.verbose = True
+        if options.wrapper:
+            print('Must not specify both a wrapper and gdb at the same time.')
+            return 1
 
     options.wd = os.path.abspath(options.wd)
 
+    if not options.no_rebuild:
+        if not rebuild_all(options.wd):
+            sys.exit(-1)
+
     th = TestHarness(options)
     if options.list:
-        return th.run_special()
-    if not options.no_rebuild:
-        if not th.rebuild_all():
-            sys.exit(-1)
+        list_tests(th)
+        return 0
     if len(options.args) == 0:
         return th.doit()
     return th.run_special()
