@@ -1,7 +1,11 @@
 #include<boost/thread.hpp>
 
+boost::recursive_mutex m;
+
 struct callable {
-    void operator()() {};
+    void operator()() {
+        boost::recursive_mutex::scoped_lock l(m);
+    };
 };
 
 int main(int argc, char **argv) {
